"""`__main__.py` is an entry point for `python -m ...`."""

import argparse
import sys
from attrs import evolve
from io import TextIOWrapper
from pathlib import Path
from itertools import chain, groupby

from . import ZapReport


DEFAULT_ALERTS_EXCLUDED = [
    10109,  # Modern Web Application
]


def main():
    """This callable is for more CLI-friendliness;
    ref: `project.scripts` at `pyproject.toml`."""

    parser = argparse.ArgumentParser(
        prog=sys.modules[__name__].__package__,
        usage='{ %(prog)s | python -m %(prog)s } [options]',
    )
    parser.add_argument(
        'in_file',
        nargs='*',
        type=argparse.FileType('r'),
        default=[sys.stdin,],
        help='Input file to parse as ZAP(-like) report, defaults to `-` (STDIN data).'
    )
    parser.add_argument(
        '-o', '--out_file',
        type=argparse.FileType('w'),
        default=None,
        help='Output file to write ZAP[-like] report to.'
             ' Defaults to STDOUT when reading from STDIN,'
             ' and to "<filename>-m.<ext>" when "<filename>.<ext>" specified as input.'
    )
    parser.add_argument(
        '-x',
        action='append',
        default=list(),
        help="Exclude alert(s) by their ZAP's `pluginid`, can be specified multiple times."
             ' Ref: <https://www.zaproxy.org/docs/alerts/>.'
             f' Use of this option overrides the default {DEFAULT_ALERTS_EXCLUDED}.'
    )
    parser.add_argument(
        '-k', '--keep-data-full', '--keep_data_full',
        action='store_true',
        help='Skip the default clearing request-response for alert instances,'
             ' except the last one within each alert.'
    )
    parser.add_argument(
        '-z', '--zap-original-output', '--zap_original_output',
        action='store_true',
        help='Use ZAP original field naming in JSON output, trying to resemble `traditional-json-plus`. Defaults to False, causing our ZAP-like output.'
             ' WARN: Mind some irreversible transformations performed:'
             ' HTML-tags removed from some fields;'
             ' some fields casted to int.'
    )
    args = parser.parse_args()

    zrs = list()
    for input_file in args.in_file:
        zr = ZapReport.from_json_file(input_file)

        while len(zr.site) > 1:
            _ = zr.site.pop(0)

        # Exclude some alerts
        zr.site[0].alerts = list(filter(
            lambda a: int(a.pluginid) not in (args.x or DEFAULT_ALERTS_EXCLUDED),
            zr.site[0].alerts
        ))

        zrs.append(zr)

    zr_merged = evolve(
        zrs[0],
        **({'program_name': '(combo)', 'version': ''} if len(zrs) > 1 else {}),
        site=[evolve(
            zrs[0].site[0],
            alerts=list(),
        ),]
    )
    kf = lambda a: (-int(a.riskcode), a.pluginid, a.alert, a.name, a.otherinfo)
    for _gk, agrp in groupby(sorted((a for zr in zrs for a in zr.site[0].alerts), key=kf), key=kf):
        agrp = list(agrp)
        ais = sorted(
            set(chain.from_iterable(a.instances for a in agrp)),
            # Keep alert instances with non-empty request/response in the end, to be consistent with further clearing
            key=lambda ai: (
                bool(ai.request_header or ai.request_body or ai.response_header or ai.response_body),
                ai,
            )
        )
        if not args.keep_data_full:
            # Clear the request/response for all but the last one
            for i in range(len(ais) - 1):
                ais[i].request_header = ''
                ais[i].request_body = ''
                ais[i].response_header = ''
                ais[i].response_body = ''
        new_alert = evolve(agrp[0], instances=ais, count=len(ais))
        zr_merged.site[0].alerts.append(new_alert)

    output_file = args.out_file
    if output_file is None:
        if args.in_file[0].name == '<stdin>':
            output_file = sys.stdout
        else:
            first_input_file = Path(args.in_file[0].name)
            output_file = first_input_file.with_stem(f'{first_input_file.stem}-m')

    with (output_file if isinstance(output_file, TextIOWrapper) else open(output_file, 'w')) as fo:
        fo.write(zr_merged.json_orig() if args.zap_original_output else zr_merged.json())


if __name__ == '__main__':
    main()



"""OWASP® Zed Attack Proxy (ZAP) reporting utilities."""

__all__ = [
    'ZapReport', 'ZapSite', 'ZapAlertInfo', 'ZapAlertInstance',
]
__version__ = '0.3'

import sys

if sys.version_info.minor < 3 or sys.version_info.minor < 9:
    raise Exception('Python >= 3.9 please.')

from .zreprt import ZapReport, ZapSite, ZapAlertInfo, ZapAlertInstance



"""
Structures here resemble OWASP® ZAP Traditional JSON Report,
with or without Requests and Responses.
Alerts are considered grouped by their info.

Changes to the Traditional JSON Report format:
- some fields renamed, keeping original names as aliases;
- some (re)typing: timestamps are ISO-formatted,
    some int and bool instead of strings;
- html tags are stripped from some fields containing descriptions.

See also:
- https://www.zaproxy.org/docs/desktop/addons/report-generation/report-traditional-json/
- https://www.zaproxy.org/docs/constants/
"""

import re
from datetime import datetime, timezone
from io import TextIOWrapper
from typing import Optional

import dateutil.parser
from attrs import astuple, define, field
from cattrs import Converter
from cattrs.gen import make_dict_structure_fn, make_dict_unstructure_fn, override
from cattrs.preconf.json import make_converter


def _clns(s, p=re.compile(r'</?p>')):
    """Clear single string of extra html tags."""
    return p.sub('', s)


_zlike_conv = make_converter()
# _zlike_conv.register_unstructure_hook(datetime, lambda dt: dt.isoformat())  # cattrs.preconf.json does this
_zlike_conv.register_structure_hook(datetime, lambda s, _: ts if (ts := dateutil.parser.parse(s)).tzinfo
                                                              else ts.replace(tzinfo=timezone.utc))
_zorig_conv = _zlike_conv.copy()


def _fallback_field(
    old_to_new_field: dict[str, str],
    zap_like_report_converter: Converter = _zlike_conv,
    zap_orig_report_converter: Converter = _zorig_conv,
):
    """Ref: https://catt.rs/en/stable/usage.html#using-fallback-key-names"""
    def decorator(cls):
        struct = make_dict_structure_fn(cls, zap_like_report_converter)

        def structure(d, cl):
            # if set(d.keys()) & set(old_to_new_field.keys()):
            for old_field, new_field in old_to_new_field.items():
                if old_field in d:
                    d[new_field] = d[old_field]
            return struct(d, cl)

        zap_like_report_converter.register_structure_hook(cls, structure)

        unstruct = make_dict_unstructure_fn(
            cls,
            zap_orig_report_converter,
            **{
                new_field: override(rename=old_field) for old_field, new_field in old_to_new_field.items()
            },
        )
        zap_orig_report_converter.register_unstructure_hook(cls, unstruct)

        return cls
    return decorator


@_fallback_field({
    "request-header": "request_header",
    "request-body": "request_body",
    "response-header": "response_header",
    "response-body": "response_body",
})
@define(order=True)
class ZapAlertInstance:
    uri: str
    method: str
    param: str
    attack: str
    evidence: str
    otherinfo: Optional[str] = ''
    request_header: Optional[str] = field(default=None, repr=False)
    request_body: Optional[str] = field(default=None, repr=False)
    response_header: Optional[str] = field(default=None, repr=False)
    response_body: Optional[str] = field(default=None, repr=False)

    def __hash__(self):
        return hash(astuple(self))


@_fallback_field({
    "alertRef": "alertref",
    "desc": "description",
})
@define
class ZapAlertInfo:
    pluginid: int
    alertref: str
    alert: str
    name: str
    riskcode: int
    confidence: int
    riskdesc: str
    description: str = field(converter=_clns)
    solution: str = field(converter=_clns)
    otherinfo: str = field(converter=_clns)
    reference: str = field(converter=_clns)
    cweid: int = field(converter=lambda v: v or -1)
    wascid: int = field(converter=lambda v: v or -1)
    sourceid: int = field(converter=lambda v: v or -1)
    instances: list[ZapAlertInstance]
    count: Optional[int] = None
    tags: Optional[list] = field(factory=list)


@_fallback_field({
    "@name": "name",
    "@host": "host",
    "@port": "port",
    "@ssl": "ssl",
})
@define
class ZapSite:
    name: str
    host: str
    port: str
    ssl: bool
    alerts: list[ZapAlertInfo]


@_fallback_field({
    "@programName": "program_name",
    "@version": "version",
    "@generated": "generated_ts",
})
@define
class ZapReport:
    """Represents ZAP Traditional JSON Report."""

    program_name: str = ''
    version: str = field(default='x3')
    generated_ts: datetime = field(factory=lambda: datetime.now(timezone.utc))
    site: list[ZapSite] = field(factory=list)

    @classmethod
    def from_json_file(cls, f):
        with (f if isinstance(f, TextIOWrapper) else open(f)) as fo:
            return _zlike_conv.loads(fo.read(), cls)

    @classmethod
    def from_dict(cls, d):
        return _zlike_conv.structure(d, cls)

    def json(self):
        return _zlike_conv.dumps(self, indent=4, ensure_ascii=False)

    def json_orig(self):
        return _zorig_conv.dumps(self, indent=4, ensure_ascii=False)
